///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Contains a simple container class.
 *	\file		IceContainer.h
 *	\author		Pierre Terdiman
 *	\date		February, 5, 2000
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Include Guard
#ifndef __ICECONTAINER_H__
#define __ICECONTAINER_H__

	class ICECORE_API Container
	{
		public:
		// Constructor / Destructor
										Container();
										~Container();

		// Management
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/**
		 *	A method to add a value in the container. The container is automatically resized if needed.
		 *	\param		entry		a udword to store in the container
		 *	\see		Add(float entry)
		 *	\see		Empty()
		 *	\see		Contains(udword entry)
		 *	\return		Self-Reference
		 */
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		__forceinline	Container&		Add(udword entry)
						{
							// Resize if needed
							if(mCurNbEntries==mMaxNbEntries)	Resize();

							// Add new entry
							mEntries[mCurNbEntries++]	= entry;
							return *this;
						}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/**
		 *	A method to add a value in the container. The container is automatically resized if needed.
		 *	\param		entry		a float to store in the container
		 *	\see		Add(udword entry)
		 *	\see		Empty()
		 *	\see		Contains(udword entry)
		 *	\return		Self-Reference
		 */
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		__forceinline	Container&		Add(float entry)
						{
							// Resize if needed
							if(mCurNbEntries==mMaxNbEntries)	Resize();

							// Add new entry
							mEntries[mCurNbEntries++]	= IR(entry);
							return *this;
						}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/**
		 *	A method to clear the container. All stored values are deleted.
		 *	\see		Reset()
		 *	\return		Self-Reference
		 */
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		__forceinline	Container&		Empty()
						{
							DELETEARRAY(mEntries);
							mCurNbEntries = mMaxNbEntries = 0;
							return *this;
						}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/**
		 *	A method to reset the container. Stored values are discarded but the buffer is kept.
		 *	\see		Empty()
		 *	\return		Self-Reference
		 */
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		__forceinline	Container&		Reset()
						{
							mCurNbEntries = 0;
							return *this;
						}

		// A method to check whether the container already contains a given value.
						bool			Contains(udword entry, udword* location=null);
		// A method to delete an entry.
						bool			Delete(udword entry);
		// A method to delete the very last entry.
		__forceinline	Container&		DeleteLastEntry()		{ if(mCurNbEntries)	mCurNbEntries--;	return *this;	}

		// Helpers
						Container&		FindNext(udword& entry, bool wrap=false);
						Container&		FindPrev(udword& entry, bool wrap=false);

		// Processing
						bool			Foreach(ENUMERATION proc, udword param=0, udword context=0);

		// Stats
						udword			GetUsedRam();

		// Data access.
		__forceinline	udword			GetNbEntries()			{ return mCurNbEntries;	}	//!< Returns the current number of entries.
		__forceinline	udword			GetEntry(udword i)		{ return mEntries[i];	}	//!< Returns ith entry
		__forceinline	udword*			GetEntries()			{ return mEntries;		}	//!< Returns the list of entries.

		private:
		// Resizing
						bool			Resize();
		// Data
						udword			mMaxNbEntries;		//!< Maximum possible number of entries
						udword			mCurNbEntries;		//!< Current number of entries
						udword*			mEntries;			//!< List of entries
	};


	class ICECORE_API LIFOStack
	{
		public:
		// Constructor/Destructor
										LIFOStack()		{}
										~LIFOStack()	{}
		// Management
		__forceinline	LIFOStack&		Push(udword entry)		{	mStack.Add(entry);	return *this;	}
						bool			Pop(udword &entry);

						udword			GetNbEntries()			{	return mStack.GetNbEntries();		}
						LIFOStack&		Reset()					{	mStack.Reset();	return *this;		}
		private:
						Container		mStack;				//!< The LIFO stack
	};

	class ICECORE_API FIFOStack
	{
		public:
		// Constructor/Destructor
										FIFOStack() : mCurIndex(0)		{}
										~FIFOStack()					{}
		// Management
		__forceinline	FIFOStack&		Push(udword entry)		{	mStack.Add(entry);	return *this;	}
						bool			Pop(udword &entry);

						udword			GetNbEntries()			{	return mStack.GetNbEntries();		}
						FIFOStack&		Reset()					{	mStack.Reset();	return *this;		}
		private:
						udword			mCurIndex;			//!< Current index within the container
						Container		mStack;				//!< The FIFO stack
	};

	// Use that one instead of an array of bools. Takes less ram, nearly as fast [no bounds checkings and so on].
	#define BITSTOBYTES(x)				((udword(x)>>3)+ ((x&7) ? 1 : 0))

	static ubyte BitMasks[]		= { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
	static ubyte NegBitMasks[]	= { 0xFE, 0xFD, 0xFB, 0xF7, 0xEF, 0xDF, 0xBF, 0x7F };

	class ICECORE_API ArrayOfBits
	{
		public:
		// Constructor / Destructor
										ArrayOfBits(udword n)		{ mBits = new ubyte[BITSTOBYTES(n)];						}			// Get ram for n bits
										~ArrayOfBits()				{ DELETEARRAY(mBits);										}			// Free used ram

		// Data management
		__forceinline	void			SetBit(udword bitnumber)	{ mBits[bitnumber>>3] |= BitMasks[bitnumber&7];				}			// Set bit bitnumber to 1
		__forceinline	void			ClearBit(udword bitnumber)	{ mBits[bitnumber>>3] &= NegBitMasks[bitnumber&7];			}			// Set bit bitnumber to 0

		// Data access
		__forceinline	bool			IsSet(udword bitnumber)		{ return (mBits[bitnumber>>3]&BitMasks[bitnumber&7])!=0;	}			// Returns true if bit bitnumber is 1

						ubyte*			mBits;																								// Array of bits
	};

#endif // __ICECONTAINER_H__
