///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Contains code for ICE kernel.
 *	\file		IceKernel.h
 *	\author		Pierre Terdiman
 *	\date		April, 4, 2000
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Include Guard
#ifndef __ICEKERNEL_H__
#define __ICEKERNEL_H__

	// Returns the V-Table [handle with care]
	ICECORE_API VTABLE		GetVTable(Cell* cell);

	// Create a dynamic identifier
	ICECORE_API	DynID		CreateID(String& desc);
	// Get the dynamic identifier descriptor
	ICECORE_API	String*		GetDescriptor(DynID id);

	// Greasy KID stuff ....
	#define	PTR(kid)				Translate(kid)
	#define	INVALID(kid)			(udword(Translate(kid))==INVALID_ID)
	#define	VALID(kid)				(udword(Translate(kid))!=INVALID_ID)
	#define CREATECLASS(classname)	GetKernel()->CreateClass(classname)

	//! Reverse all the bits in a 32 bit word (from Steve Baker's Cute Code Collection)
	__forceinline void ReverseBits(udword& n)
	{
		n = ((n >>  1) & 0x55555555) | ((n <<  1) & 0xaaaaaaaa) ;
		n = ((n >>  2) & 0x33333333) | ((n <<  2) & 0xcccccccc) ;
		n = ((n >>  4) & 0x0f0f0f0f) | ((n <<  4) & 0xf0f0f0f0) ;
		n = ((n >>  8) & 0x00ff00ff) | ((n <<  8) & 0xff00ff00) ;
		n = ((n >> 16) & 0x0000ffff) | ((n << 16) & 0xffff0000) ;
	}

	//! Count the number of '1' bits in a 32 bit word (from Steve Baker's Cute Code Collection)
	__forceinline void CountBits(udword& n)
	{
		n = (n & 0x55555555) + ((n & 0xaaaaaaaa) >> 1);
		n = (n & 0x33333333) + ((n & 0xcccccccc) >> 2);
		n = (n & 0x0f0f0f0f) + ((n & 0xf0f0f0f0) >> 4);
		n = (n & 0x00ff00ff) + ((n & 0xff00ff00) >> 8);
		n = (n & 0x0000ffff) + ((n & 0xffff0000) >> 16);
	}

	//! Test to see if a number is an exact power of two (from Steve Baker's Cute Code Collection)
	__forceinline bool IsPowerOfTwo(udword n)				{ return ((n&(n-1))==0);			}

	//! Zero the least significant '1' bit in a word. (from Steve Baker's Cute Code Collection)
	__forceinline void ZeroLeastSetBit(udword& n)			{ n&=(n-1);							}

	//! Set the least significant N bits in a word. (from Steve Baker's Cute Code Collection)
	__forceinline void SetLeastNBits(udword& x, udword n)	{ x|=~(~0<<n);						}

	//! Classic XOR swap (from Steve Baker's Cute Code Collection)
	__forceinline void Swap(udword& x, udword& y)			{ x ^= y; y ^= x; x ^= y;			}

	//! Little/Big endian (from Steve Baker's Cute Code Collection)
	__forceinline char LittleEndian()						{ int i = 1; return *((char*)&i);	}

	//! Returns a unit random floating-point value
	__forceinline float UnitRandomFloat()					{ return float(rand()) * ONE_OVER_RAND_MAX;	}

	__forceinline udword CodeSize(udword value)
	{
		int n;
		*((float*)&n) = (float)value;
		n = (n>>23)-126;
		return n;
/*		udword NbBits = 0;
		for(udword i=0;i<32;i++)
		{
			udword Limit = 1<<i;
			if(value>Limit)	NbBits++;
			else break;
		}
		return NbBits;*/
	}

	// Returns the global kernel
	ICECORE_API	Kernel*	GetKernel();

	// Translation shortcut
	ICECORE_API	Cell*	Translate(KID kid);

	//! Kernel flags
	enum KernelFlag
	{
		KF_CELL_CONTROL		=	(1<<0),		//!< Enable high-level cell control (self-destructions, etc)
		KF_MESSAGES			=	(1<<1),		//!< Enable global messages
		KF_SECURITY			=	(1<<2),		//!< Enable high-level security. Kernel is allowed to automatically disable "explosive" features.
		KF_CELL_NOTIF		=	(1<<3),		//!< Enable Kernel-to-cells notifications
		KF_CELL_MODIF		=	(1<<4),		//!< Allow the kernel to automatically fix bad pointers within cells (handle that one with care)
		KF_FORCE_DWORD		=	0x7fffffff
	};

	class ICECORE_API Kernel
	{
		public:
		// Constructor/Destructor
										Kernel();
										~Kernel();

						bool			RegisterAPI		(const char* api, CLASSCREATOR callback);
						Cell*			CreateClass		(const char* classname);

						bool			Register		(Cell* object);
						bool			Unregister		(Cell* object);

		__forceinline	bool			CreateRef		(Cell* owner, Cell* ref)	{ return mTracker.AddRef((udword)owner->m__KernelID, (udword)ref->m__KernelID);}
		__forceinline	bool			CreateRef		(KID ownerkid, KID refkid)	{ return mTracker.AddRef((udword)ownerkid, (udword)refkid);	}
						bool			DeleteRef		(Cell* owner, Cell* ref);
						bool			DeleteRef		(KID ownerkid, KID refkid);

						bool			ReplaceRef		(Cell* owner, Cell* oldref, Cell* newref);

		__forceinline	udword			GetNbRefs		(Cell* cell)				{ return mTracker.GetNbRefs((udword)cell->m__KernelID);		}
		__forceinline	udword			GetNbOwners		(Cell* cell)				{ return mTracker.GetNbOwners((udword)cell->m__KernelID);	}
						Cell*			GetOwner		(Cell* cell, udword i);
						Cell*			GetRef			(Cell* cell, udword i);

		__forceinline	bool			CreateTypedRef	(Cell* owner, Cell* ref, RTYPE type)	{ return mTracker.AddRef((udword)owner->m__KernelID, (udword)ref->m__KernelID, type);}

						udword			GetCells		(Container& list, udword type);

		// Check kernel integrity
						bool			Checkup			(bool deletelostcells=false);

		//! Cell access. The first entry is always the kernel itself, which is not even a cell-based component.
		//! Hence the -1 and +1 to return the relevant parts only.
		__forceinline	udword			GetNbEntries	()					{ return mTranslator.GetNbEntries()-1;		}
		__forceinline	udword			GetMaxNbEntries	()					{ return mTranslator.GetMaxNbEntries()-1;	}
		__forceinline	Cell**			GetEntries		()					{ return ((Cell**)mTranslator.GetList())+1;	}

		__forceinline	udword			GetTrackerRam	()					{ return mTracker.GetUsedRam();				}

		//! High-speed KernelID-to-cell translation
		__forceinline	Cell*			Translate		(KID kid)			{ return (Cell*)mTranslator.GetValue(kid);	}

		// Flags
		__forceinline	uword			GetFlags		()					{ return mFlags;							}
		__forceinline	bool			IsSet			(KernelFlag flag)	{ return (mFlags&flag)!=0;					}
		__forceinline	Kernel&			Enable			(KernelFlag flag)	{ mFlags |= flag;	return *this;			}
		__forceinline	Kernel&			Disable			(KernelFlag flag)	{ mFlags &= ~flag;	return *this;			}

		//! Cell self-destruction user-param
		__forceinline	Kernel&			SetSDUP			(udword param)		{ mSDUP = param;	return *this;			}

		// Debug
						RefTracker*		GetTracker()						{ return &mTracker; }
		private:
						udword			mSDUP;				//!< Self-Destruction user-parameter
						udword			mFlags;				//!< Kernel flags
						RefTracker		mTracker;			//!< Reference tracker
						DynamicList		mTranslator;		//!< Main translation module
						Container		mCallbacks;			//!< Class creators

		// Secure cell self-destruction
						Kernel&			TerminateCell	(Cell* cell);
		// Automatic reference fixing - that one is dangerous and should be kept private...
						bool			FixInvalidReference(Cell* owner, Cell* badref, Cell* newref=null, bool allowmodif=true);
	};

	class ICECORE_API SmartKID
	{
		public:
										SmartKID(KID kid=INVALID_KID)	: mKid(kid)		{}
										~SmartKID()										{}

										operator void*()		{ return (void*)(udword(Translate(mKid))-INVALID_ID);	}
						Cell*			operator->()	const	{ return PTR(mKid);										}

//						SmartKID& operator= (const SmartKID& k)	{ return *this; }
//						Cell&		operator*()		const	{ return *PTR(mKid); }

		private:
						KID				mKid;

				friend	class			Kernel;
	};

#endif // __ICEKERNEL_H__
